/*
 * crash-manager
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <limits.h>
#include <unistd.h>
#include <libgen.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/prctl.h>
#include <sys/file.h>
#include <sys/vfs.h>
#include <gio/gio.h>
#include <iniparser.h>
#include <tzplatform_config.h>
#include "crash-manager.h"
#include "shared/log.h"
#include "shared/util.h"

#undef LOG_TAG
#define LOG_TAG "CRASH_MANAGER"

#define DEBUGMODE_FILE       tzplatform_mkpath(TZ_SYS_ETC, ".debugmode")

/* Parsing */
#define CRASH_CONF_FILE      tzplatform_mkpath(TZ_SYS_ETC, "crash-manager.conf")
#define KEY_MAX              255
#define CRASH_SECTION        "CrashManager"

/* Crash-popup dbus */
#define POPUP_BUS_NAME       "org.tizen.system.popup"
#define POPUP_OBJECT_PATH    "/Org/Tizen/System/Popup/Crash"
#define POPUP_INTERFACE_NAME POPUP_BUS_NAME".Crash"
#define POPUP_METHOD         "PopupLaunch"

/* Configuration default values */
/* note: 0 means unlimited */
#define SYSTEM_MAX_USE       0
#define SYSTEM_KEEP_FREE     0
#define MAX_RETENTION_SEC    0
#define MAX_CRASH_DUMP       0
#define ALLOW_ZIP            true

#define CRASH_CHECK_DISK_PATH "/opt/usr"

enum {
	RET_EXCEED = 1,
	NUM_EXCEED,
	USAGE_EXCEED
};

struct file_info {
	bool   isdir;
	int    size;
	time_t mtime;
	char   *path;
};

/* Configuration variables */
static int system_max_use;
static int system_keep_free;
static int max_retention_sec;
static int max_crash_dump;
static bool allow_zip;

/* Paths and variables */
static struct crash_info {
	char *cmd_info;
	char *pid_info;
	char *tid_info;
	char *sig_info;
	char time_info[80];
	char temp_dir[PATH_MAX];
	char name[FILENAME_MAX];
	char result_path[PATH_MAX];
	char pfx[PATH_MAX];
	char info_path[PATH_MAX];
	char core_path[PATH_MAX];
	char log_path[PATH_MAX];
#ifdef SYS_ASSERT
	char sysassert_cs_path[PATH_MAX];
#endif
} crash_info;

static void get_config(void)
{
	dictionary *ini = NULL;
	char key[KEY_MAX];
	int value;

	system_max_use = SYSTEM_MAX_USE;
	system_keep_free = SYSTEM_KEEP_FREE;
	max_retention_sec = MAX_RETENTION_SEC;
	max_crash_dump = MAX_CRASH_DUMP;
	allow_zip = ALLOW_ZIP;

	ini = iniparser_load(CRASH_CONF_FILE);
	if (!ini) {
		_E("Failed to load conf file");
		return;
	}

	snprintf(key, sizeof(key), "%s:%s", CRASH_SECTION, "SystemMaxUse");
	value = iniparser_getint(ini, key, -1);
	if (value < 0) {
		_D("Invalid value for SystemMaxUse. Use default value [ %d kbyte]",
				SYSTEM_MAX_USE);
	} else {
		_D("SystemMaxUse [ %d kbyte]", value);
		system_max_use = value;
	}

	snprintf(key, sizeof(key), "%s:%s", CRASH_SECTION, "SystemKeepFree");
	value = iniparser_getint(ini, key, -1);
	if (value < 0) {
		_D("Invalid value for SystemKeepFree. Use default value [ %d kbyte]",
				SYSTEM_KEEP_FREE);
	} else {
		_D("SystemKeepFree [ %d kbyte]", value);
		system_keep_free = value;
	}


	snprintf(key, sizeof(key), "%s:%s", CRASH_SECTION, "MaxRetentionSec");
	value = iniparser_getint(ini, key, -1);
	if (value < 0) {
		_D("Invalid value for MaxRetentionSec. Use default value [ %d ]",
				MAX_RETENTION_SEC);
	} else {
		_D("MaxRetentionSec [ %d ]", value);
		max_retention_sec = value;
	}

	snprintf(key, sizeof(key), "%s:%s", CRASH_SECTION, "MaxCrashDump");
	value = iniparser_getint(ini, key, -1);
	if (value < 0) {
		_D("Invalid value for MaxCrashDump. Use default value [ %d ]",
				MAX_CRASH_DUMP);
	} else {
		_D("MaxCrashDump [ %d ]", value);
		max_crash_dump = value;
	}

	snprintf(key, sizeof(key), "%s:%s", CRASH_SECTION, "AllowZip");
	value = iniparser_getboolean(ini, key, -1);
	if (value < 0) {
		_D("Invalid value for AllowZip. Use default value [ %s ]",
				ALLOW_ZIP ? "true" : "false" );
	} else {
		_D("AllowZip [ %s ]", value ? "true" : "false");
		allow_zip = value;
	}

	iniparser_freedict(ini);
}

static int make_dump_dir(void)
{
	struct stat st;

	if (!stat(CRASH_PATH, &st)) {
		if (!(st.st_mode & S_IFDIR)) {
			_E("%s (not DIR) is already exist", CRASH_PATH);
			return -1;
		}
	} else {
		if (mkdir(CRASH_PATH, 0775) < 0) {
			_E("Failed to mkdir for %s", CRASH_PATH);
			return -1;
		}
	}

	if (!stat(CRASH_TEMP, &st)) {
		if (!(st.st_mode & S_IFDIR)) {
			_E("%s (not DIR) is already exist", CRASH_TEMP);
			return -1;
		}
	} else {
		if (mkdir(CRASH_TEMP, 0775) < 0) {
			_E("Failed to mkdir for %s", CRASH_TEMP);
			return -1;
		}
	}

	return 0;
}

static int set_crash_info(int argc, char *argv[])
{
	int ret;
	char *temp_dir_ret;
	time_t time_val;
	struct tm loc_tm;

	crash_info.pid_info = argv[1];
	crash_info.sig_info = argv[4];
	crash_info.cmd_info = argv[6];
	if (argc > 8)
		crash_info.tid_info = argv[8];

	time_val = atoll(argv[5]);
	localtime_r(&time_val, &loc_tm);
	strftime(crash_info.time_info, sizeof(crash_info.time_info),
			"%Y%m%d%H%M%S", &loc_tm);

	ret = snprintf(crash_info.temp_dir, sizeof(crash_info.temp_dir),
			"%s/crash.XXXXXX", CRASH_TEMP);
	if (ret < 0) {
		_E("Failed to snprintf for temp_dir");
		return -1;
	}
	temp_dir_ret = mkdtemp(crash_info.temp_dir);
	if (access(temp_dir_ret, F_OK)) {
		_E("Failed to mkdtemp for temp_dir");
		return -1;
	}

	ret = snprintf(crash_info.name, sizeof(crash_info.name), "%s_%s_%s",
			crash_info.cmd_info,
			crash_info.pid_info,
			crash_info.time_info);
	if (ret < 0) {
		_E("Failed to snprintf for name");
		goto rm_temp;
	}

	if (allow_zip)
		ret = snprintf(crash_info.result_path,
				sizeof(crash_info.result_path),
				"%s/%s.zip", CRASH_PATH, crash_info.name);
	else
		ret = snprintf(crash_info.result_path,
				sizeof(crash_info.result_path),
				"%s/%s", CRASH_PATH, crash_info.name);
	if (ret < 0) {
		_E("Failed to snprintf for result path");
		goto rm_temp;
	}

	ret = snprintf(crash_info.pfx, sizeof(crash_info.pfx), "%s/%s",
			crash_info.temp_dir, crash_info.name);
	if (ret < 0) {
		_E("Failed to snprintf for pfx");
		goto rm_temp;
	}
	ret = mkdir(crash_info.pfx, 0775);
	if (ret < 0) {
		_E("Failed to mkdir for %s", crash_info.pfx);
		goto rm_temp;
	}

	ret = snprintf(crash_info.info_path, sizeof(crash_info.info_path),
			"%s/%s.info", crash_info.pfx, crash_info.name);
	if (ret < 0) {
		_E("Failed to snprintf for info path");
		goto rm_temp;
	}

	ret = snprintf(crash_info.core_path, sizeof(crash_info.core_path),
			"%s/%s.coredump", crash_info.pfx, crash_info.name);
	if (ret < 0) {
		_E("Failed to snprintf for core path");
		goto rm_temp;
	}

	ret = snprintf(crash_info.log_path, sizeof(crash_info.log_path),
			"%s/%s.log", crash_info.pfx, crash_info.name);
	if (ret < 0) {
		_E("Failed to snprintf for log path");
		goto rm_temp;
	}

#ifdef SYS_ASSERT
	ret = snprintf(crash_info.sysassert_cs_path,
			sizeof(crash_info.sysassert_cs_path),
		       "/tmp/crash_stack/%s_%s.info",
		       crash_info.cmd_info, crash_info.pid_info);
	if (ret < 0) {
		_E("Failed to snprintf for sys-assert callstack path");
		goto rm_temp;
	}
#endif

	return 0;

rm_temp:
	remove_dir(crash_info.temp_dir, 1);
	return -1;
}

#ifdef SYS_ASSERT
static int get_sysassert_cs(void)
{
	int ret;
	char move_path[PATH_MAX];

	if (access(crash_info.sysassert_cs_path, F_OK)) {
		_E("The sys-assert cs file not found: %s",
			crash_info.sysassert_cs_path);
		return -1;
	}

	ret = snprintf(move_path, sizeof(move_path), "%s/%s",
			crash_info.pfx, basename(crash_info.sysassert_cs_path));
	if (ret < 0) {
		_E("Failed to snprintf for move path");
		return -1;
	}

	if (move_file(crash_info.sysassert_cs_path, move_path) < 0) {
		_E("Failed to move %s to %s",
				crash_info.sysassert_cs_path, move_path);
		return -1;
	}

	return 0;
}
#endif

static int convert_path(char *path, char *core_exepath)
{
	int i;

	for (i = 0; core_exepath[i]; i++)
		path[i] = (core_exepath[i] == '!' ? '/' : core_exepath[i]);
	path[i] = '\0';

	return i;
}

static void launch_crash_popup(char *core_exepath)
{
	GDBusConnection *conn;
	GVariantBuilder *builder;
	GVariant *parameters = NULL;
	GVariant *reply = NULL;
	GError *error = NULL;
	int ret;
	char exepath[PATH_MAX] = "\0";

	if (convert_path(exepath, core_exepath) <= 0) {
		_E("Failed to parsing exepath");
		return;
	}

	conn = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &error);
	if (error) {
		_E("Failed to get dbus: %s", error->message);
		g_error_free(error);
		return;
	}

	builder = g_variant_builder_new(G_VARIANT_TYPE("a{ss}"));
	g_variant_builder_add(builder, "{ss}", "_SYSPOPUP_CONTENT_", "crash");
	g_variant_builder_add(builder, "{ss}", "_PROCESS_NAME_",
			crash_info.cmd_info);
	g_variant_builder_add(builder, "{ss}", "_EXEPATH_", exepath);
	parameters = g_variant_new("(a{ss})", builder);
	g_variant_builder_unref(builder);

	reply = g_dbus_connection_call_sync(conn,
					    POPUP_BUS_NAME,
					    POPUP_OBJECT_PATH,
					    POPUP_INTERFACE_NAME,
					    POPUP_METHOD,
					    parameters,
					    G_VARIANT_TYPE("(i)"),
					    G_DBUS_CALL_FLAGS_NONE,
					    120000,
					    NULL,
					    &error);
	if (error) {
		_E("Failed to get reply: %s", error->message);
		g_error_free(error);
		goto exit;
	}

	g_variant_get(reply, "(i)", &ret);
	_I("Crash_popup is launched: (%d)", ret);

exit:
	if (reply)
		g_variant_unref(reply);
	if (parameters)
		g_variant_unref(parameters);
}

static void dump_system_state(void)
{
	int ret;
	char command[PATH_MAX];

	ret = snprintf(command, sizeof(command),
			"/usr/bin/dump_systemstate -d -k -f %s",
			crash_info.log_path);
	if (ret < 0) {
		_E("Failed to snprintf for dump_systemstate command");
		return;
	}
	system_command(command);
}

static void execute_crash_modules(int argc, char *argv[], int debug)
{
	int ret;
	char command[PATH_MAX];

	/* Execute crash-pipe */
	if (debug) {
		ret = snprintf(command, sizeof(command),
				"%s --save-core %s",
				CRASH_PIPE_PATH,
				crash_info.core_path);
		if (ret < 0) {
			_E("Failed to snprintf for crash-pipe command");
			return;
		}
		system_command(command);
	}

#ifdef TIZEN_FEATURE_PTRACE_CALLSTACK
	/* Execute crash-stack */
	if (argc > 8)
		ret = snprintf(command, sizeof(command),
				"%s --pid %s --tid %s --sig %s > %s",
				CRASH_STACK_PATH,
				crash_info.pid_info,
				crash_info.tid_info,
				crash_info.sig_info,
				crash_info.info_path);
	else
		ret = snprintf(command, sizeof(command),
				"%s --pid %s --sig %s > %s",
				CRASH_STACK_PATH,
				crash_info.pid_info,
				crash_info.sig_info,
				crash_info.info_path);
	if (ret < 0) {
		_E("Failed to snprintf for crash-stack command");
		return;
	}
	system_command(command);
#endif /* TIZEN_FEATURE_PTRACE_CALLSTACK */
}

static int lock_dumpdir(void)
{
	int fd;

	if ((fd = open(CRASH_PATH, O_RDONLY | O_DIRECTORY)) < 0) {
		_E("Failed to open %s", CRASH_PATH);
		return -1;
	}

	if (flock(fd, LOCK_EX) < 0) {
		_E("Failed to flock (LOCK)");
		return -1;
	}

	return fd;
}

static void unlock_dumpdir(int fd)
{
	if (flock(fd, LOCK_UN) < 0)
		_E("Failed to unlink (UNLOCK)");
	close(fd);
}

static int dump_filter(const struct dirent *de)
{
	if (de->d_name[0] == '.')
		return 0;
	return 1;
}

static int mtime_cmp(const void *_a, const void *_b)
{
	const struct file_info *a = _a;
	const struct file_info *b = _b;

	if (a->mtime < b->mtime)
		return -1;
	if (a->mtime > b->mtime)
		return 1;
	return 0;
}

static int scan_dump(struct file_info **dump_list, size_t *usage)
{
	struct file_info *temp_list;
	struct dirent **scan_list = NULL;
	struct stat st;
	int i, scan_num, dump_num = 0;
	int fd;

	if ((fd = open(CRASH_PATH, O_DIRECTORY)) < 0 ) {
		_E("Failed to open %s", CRASH_PATH);
		return -1;
	}

	scan_num = scandir(CRASH_PATH, &scan_list, &dump_filter, NULL);
	if (scan_num < 0) {
		close(fd);
		return -1;
	}

	temp_list = (struct file_info *)calloc(scan_num,
			sizeof(struct file_info));
	if (!temp_list) {
		_E("Failed to calloc for dump list");
		goto exit;
	}

	for (i = 0; i < scan_num; i++) {
		if (fstatat(fd, scan_list[i]->d_name, &st, 0) < 0) {
			_E("Failed to fstatat");
			continue;
		}

		if (asprintf(&(temp_list[dump_num].path), "%s/%s",
					CRASH_PATH, scan_list[i]->d_name) < 0) {
			_E("Failed to asprintf");
			continue;
		}

		if (scan_list[i]->d_type == DT_DIR) {
			temp_list[dump_num].isdir = 1;
			temp_list[dump_num].size =
				get_directory_usage(temp_list[dump_num].path);
		} else {
			temp_list[dump_num].isdir = 0;
			temp_list[dump_num].size = st.st_size;
		}
		temp_list[dump_num].mtime = st.st_mtime;
		dump_num++;
	}

	if (dump_num <= 0) {
		free(temp_list);
		goto exit;
	}

	if (dump_num != scan_num)
		temp_list = (struct file_info *)realloc(temp_list,
				dump_num * sizeof(struct file_info));

	qsort(temp_list, dump_num, sizeof(struct file_info), mtime_cmp);

	*usage = 0;
	for (i = 0; i < dump_num; i++) {
		_D("[%d] path: %s(%s), size: %d kb, mtime: %s",
				i,
				temp_list[i].path,
				temp_list[i].isdir ? "DIR" : "FILE",
				temp_list[i].size / 1024,
				ctime(&(temp_list[i].mtime)));
		*usage += temp_list[i].size;
	}
	*dump_list = temp_list;
exit:
	for (i = 0; i < scan_num; i++)
		free(scan_list[i]);
	free(scan_list);
	close(fd);

	return dump_num;
}

static int check_disk_available(const char *path, int check_size)
{
	struct statfs lstatfs;
	int avail_size = 0;

	if (!path)
		return -1;

	if (statfs(path, &lstatfs) < 0)
		return -1;
	avail_size = (int)(lstatfs.f_bavail * (lstatfs.f_bsize / 1024));

	if (check_size > avail_size) {
		_I("avail_size is (%d)", avail_size);
		return -1;
	}

	return 0;
}

static int remove_file(struct file_info file)
{
	if (file.isdir)
		return remove_dir(file.path, 1);
	else
		return unlink(file.path);
}

static void clean_dump(void)
{
	struct file_info *dump_list = NULL;
	int i, scan_num, dump_num, remove_flag;
	size_t usage = 0;
	time_t cur_time;

	scan_num = scan_dump(&dump_list, &usage);
	if (scan_num <= 0)
		return;

	dump_num = scan_num;

	cur_time = time(NULL);

	for (i = 0; i < scan_num; i++) {
		remove_flag = 0;

		/* Retention time check */
		if (max_retention_sec &&
			dump_list[i].mtime > 0 &&
			dump_list[i].mtime + max_retention_sec < cur_time)
			remove_flag = RET_EXCEED;
		/* Check the number of dumps */
		else if (max_crash_dump &&
			0 < dump_num && max_crash_dump < dump_num)
			remove_flag = NUM_EXCEED;
		/* Check the max system use size */
		else if (system_max_use &&
			0 < dump_num && system_max_use < usage / 1024)
			remove_flag = USAGE_EXCEED;

		switch (remove_flag) {
		case RET_EXCEED:
			_I("Reached the maximum retention time %d, so remove (%s)",
					max_retention_sec, dump_list[i].path);
			break;
		case NUM_EXCEED:
			_I("Reached the maximum number of dump %d/%d, so remove (%s)",
					dump_num, max_crash_dump,
					dump_list[i].path);
			break;
		case USAGE_EXCEED:
			_I("Reached the maximum disk usage %d/%d kb, so remove (%s)",
					usage / 1024, system_max_use,
					dump_list[i].path);
			break;
		default:
			break;
		}

		if (!remove_flag)
			continue;

		if (remove_file(dump_list[i]) < 0) {
			_E("Failed to remove %s", dump_list[i].path);
			continue;
		}
		dump_num--;
		usage -= dump_list[i].size;
	}

	/* Check disk free space to keep */
	if (system_keep_free &&
			check_disk_available(CRASH_CHECK_DISK_PATH,
					     system_keep_free) < 0) {
		_I("Disk is not available! so set the maximum number of dump to 1");
		max_crash_dump = 1;
	}

	for (i = 0; i < dump_num; i++)
		free(dump_list[i].path);
	free(dump_list);
}

static void compress(void)
{
	int ret, lock_fd;
	char zip_path[PATH_MAX];
	char command[PATH_MAX];

	ret = snprintf(zip_path, sizeof(zip_path), "%s/report.zip",
			crash_info.temp_dir);
	if (ret < 0) {
		_E("Failed to snprintf for zip path");
		return;
	}

	ret = snprintf(command, sizeof(command), "cd %s && /bin/zip -r %s %s > /dev/null 2>&1",
			crash_info.temp_dir, zip_path, crash_info.name);
	if (ret < 0) {
		_E("Failed to snprintf for zip command");
		return;
	}
	system_command(command);

	if ((lock_fd = lock_dumpdir()) < 0)
		return;
	if (!rename(zip_path, crash_info.result_path))
		clean_dump();
	else
		_E("Failed to move %s to %s", zip_path, crash_info.result_path);
	unlock_dumpdir(lock_fd);

	if (remove_dir(crash_info.temp_dir, 1) < 0)
		_E("Failed to delete temp directory");
}

static void move_dump_dir(void)
{
	int lock_fd;

	if ((lock_fd = lock_dumpdir()) < 0)
		return;
	if (!rename(crash_info.pfx, crash_info.result_path))
		clean_dump();
	else
		_E("Failed to move %s to %s",
				crash_info.pfx, crash_info.result_path);
	unlock_dumpdir(lock_fd);

	if (remove_dir(crash_info.temp_dir, 1) < 0)
		_E("Failed to delete temp directory");
}

int main(int argc, char *argv[])
{
	prctl(PR_SET_DUMPABLE, 0);

	/* Get Configuration */
	get_config();

	/* Create crash directories */
	if (make_dump_dir() < 0)
		exit(EXIT_FAILURE);

	/* Set crash info */
	if (set_crash_info(argc, argv) < 0)
		exit(EXIT_FAILURE);

#ifdef SYS_ASSERT
	/* Fetch callstack of sys-assert */
	get_sysassert_cs();
#endif

	/* .dbugmode: launch crash-popup */
	if (access(DEBUGMODE_FILE, F_OK) == 0)
		launch_crash_popup(argv[7]);

	/* Exec dump_systemstate */
	dump_system_state();

	/* Exec crash modules */
	execute_crash_modules(argc, argv, DEBUG);

	/* Tar compression */
	if (allow_zip)
		compress();
	else
		move_dump_dir();

	return 0;
}
